#include <stdio.h>
#include <stdlib.h>
#include <math.h>

// use initial data to determine exact intersections

int w[1024*8] = {
#include "world.h"
};


// in the data we are going to generate, the line info will be the y coord
// for segment i to the x,y coord for segment i+1
// ie backx[i],backy[i] to backx[i+1],backy[i+1]

// if the line is not to be drawn, backy[i] will be > max allowed value

// foreground line is always drawn, and is forex[i],forey[i] to forex[i+1],forey[i+1]

// intersection coord is calculated in real world coordinates then reversed back
// to i<<10+delta with the delta being stored.

/*
   Determine the intersection point of two line segments
   Return FALSE if the lines don't intersect
*/
static inline double ABS(double x) {
  if (x < 0.0) return -x;
  return x;
}
static double EPS=0.1;
int fintersect(
		  double x1, double y1,
		  double x2, double y2,
		  double x3, double y3,
		  double x4, double y4,
		  double *x, double *y)
{
  double mua,mub;
  double denom,numera,numerb;

  denom  = (y4-y3) * (x2-x1) - (x4-x3) * (y2-y1);
  numera = (x4-x3) * (y1-y3) - (y4-y3) * (x1-x3);
  numerb = (x2-x1) * (y1-y3) - (y2-y1) * (x1-x3);

  /* Are the line coincident? */
  if (ABS(numera) < EPS && ABS(numerb) < EPS && ABS(denom) < EPS) {
    *x = (x1 + x2) / 2;
    *y = (y1 + y2) / 2;
    return(1);
  }

  /* Are the line parallel */
  if (ABS(denom) < EPS) {
    *x = 0;
    *y = 0;
    return(0);
  }

  /* Is the intersection along the the segments */
  mua = numera / denom;
  mub = numerb / denom;
  if (mua < 0 || mua > 1 || mub < 0 || mub > 1) {
    *x = 0;
    *y = 0;
    return(0);
  }
  *x = x1 + mua * (x2 - x1);
  *y = y1 + mua * (y2 - y1);
  return(1);
}

int intersect(
		  int x1, int y1,
		  int x2, int y2,
		  int x3, int y3,
		  int x4, int y4,
		  int *x, int *y) {
  double fx,fy;
  // ints will be converted to doubles by C
  int ok = fintersect(x1,y1,x2,y2,x3,y3,x4,y4,&fx,&fy);
  if (ok) {
    *x = round(fx); *y = round(fy);
  }
  return ok;
}

int main(int argc, char **argv) {
  int bgx[1024],bgy[1024],fgx[1024],fgy[1024];
  int fromx[1024],fromy[1024]; // for partial (clipped) lines, coords relative to
                               // start of segment
  int tox[1024],toy[1024];
  int i, prev,this,next, prev_idx,idx,next_idx,
      x,y,this_fg_x,this_fg_y,next_fg_x,next_fg_y,
      this_bg_x,this_bg_y,next_bg_x,next_bg_y,
      intersect_x,intersect_y,
      this_x,this_y;
  
  for (x = 0; x < 1024; x++) {
    idx = x;  this = idx*8;

    this_fg_x = w[this+0]; this_fg_y=w[this+1];
    this_bg_x = w[this+4]; this_bg_y=w[this+5];

    fgx[idx] = this_fg_x; fgy[idx] = this_fg_y;
    bgx[idx] = this_bg_x; bgy[idx] &= 1<<30; bgy[idx] |= this_bg_y; 
  }

  // fgx and bgx both 0 in this revision to keep it simple.

  // horizon is 1024 segments of 16 ticks each
  for (x = 0; x < 1024; x++) {
    idx = x&1023; next_idx=(idx+1)&1023; prev_idx=(idx-1)&1023;
    this = idx; next = (idx+1)&1023; prev = (idx-1)&1023;
    this *= 8; next *= 8; prev *= 8;

    this_fg_x = w[this+0];    this_fg_y=w[this+1];
    next_fg_x = w[this+2]+16; next_fg_y=w[this+3];
    this_bg_x = w[this+4];    this_bg_y=w[this+5];
    next_bg_x = w[this+6]+16; next_bg_y=w[this+7];

    if ((this_bg_y <= this_fg_y) && (next_bg_y <= next_fg_y)) {
      // if bg line is fully below fg line, drop it completely
      // need to mark as 'do not draw from segment idx to idx+1'
      fromx[idx] = bgx[idx]|(1<<30);
      fromy[idx] = bgy[idx];
      tox[idx] = bgx[next_idx]+16;
      toy[idx] = bgy[next_idx];
    } else if ((this_bg_y >= this_fg_y) && (next_bg_y >= next_fg_y)) {
      // if bg line is fully above fg line, draw it
      fromx[idx] = bgx[idx];
      fromy[idx] = bgy[idx];
      tox[idx] = bgx[next_idx]+16;
      toy[idx] = bgy[next_idx];
    } else {
      //
      // if lines intersect, draw the part of the bg line that is above the fg line
      // starts where it starts. background x and y OF NEXT SEGMENT are adjusted to
      // the calculated intersection - the foreground line is never changed.
      //
      if (this_fg_y < next_fg_y) {
        // case 1: fg: /   bg: \ ...
	if (!intersect(this_fg_x,this_fg_y, next_fg_x,next_fg_y,
		  this_bg_x,this_bg_y, next_bg_x,next_bg_y,
		       &intersect_x,&intersect_y)) {
          fprintf(stderr, "No intersection?  (%d,%d):(%d,%d) X (%d,%d):(%d,%d)\n",
		  this_fg_x,this_fg_y,next_fg_x,next_fg_y,this_bg_x,this_bg_y, next_bg_x,next_bg_y);
	  exit(1);
	} else {
          fprintf(stderr, "Intersection(%4d): /\\ (%d,%d):(%d,%d) X (%d,%d):(%d,%d) at (%d,%d)\n",
		  x,
		  this_fg_x,this_fg_y,next_fg_x,next_fg_y,this_bg_x,this_bg_y, next_bg_x,next_bg_y,
		  intersect_x,intersect_y);
	}
	// start of bg line is same,
	fromx[idx] = bgx[idx];
	fromy[idx] = bgy[idx];
	// end of bg line (idx+1) is moved left and up.
	tox[idx] = intersect_x;
	toy[idx] = intersect_y;
      } else if (this_fg_y > next_fg_y) {
        // case 2: fg: \  bg: /
	if (!intersect(this_fg_x,this_fg_y, next_fg_x,next_fg_y,
		  this_bg_x,this_bg_y, next_bg_x,next_bg_y,
		       &intersect_x,&intersect_y)) {
          fprintf(stderr, "No intersection?  (%d,%d):(%d,%d) X (%d,%d):(%d,%d)\n",
		  this_fg_x,this_fg_y,next_fg_x,next_fg_y,this_bg_x,this_bg_y, next_bg_x,next_bg_y);
	  exit(1);
	} else {
          fprintf(stderr, "Intersection(%4d): \\/ (%d,%d):(%d,%d) X (%d,%d):(%d,%d) at (%d,%d)\n",
		  x,
		  this_fg_x,this_fg_y,next_fg_x,next_fg_y,this_bg_x,this_bg_y, next_bg_x,next_bg_y,
		  intersect_x,intersect_y);
	}
	// start of bg line (idx) is moved right and up,
	fromx[idx] = intersect_x;
	fromy[idx] = intersect_y;
	// end of bg line is same.
	tox[idx] = bgx[next_idx]+16;
	toy[idx] = bgy[next_idx];
      } else {
	// BUG?
	//fprintf(stderr, "BUG?\n");
	//exit(1);
	fromx[idx] = bgx[idx];
	fromy[idx] = bgy[idx];
	tox[idx] = bgx[next_idx];
	toy[idx] = bgy[next_idx];
      }
    }    
  }
#ifdef never
  fprintf(stdout, "static const int fgx[1024] = {\n");
  for (i = 0; i < 1024; i++) {
    fprintf(stdout, "  %d,", fgx[i]-((i)<<4) );
    if ((i&15)==15) fprintf(stdout, "\n");
  }
  fprintf(stdout, "};\n");
#endif
  
  fprintf(stdout, "static const int fgy[1024] = {\n");
  for (i = 0; i < 1024; i++) {
    fprintf(stdout, "  %d,", fgy[i]);
    if ((i&15)==15) fprintf(stdout, "\n");
  }
  fprintf(stdout, "};\n");
#ifdef never
  fprintf(stdout, "static const int bgx[1024] = {\n");
  for (i = 0; i < 1024; i++) {
    fprintf(stdout, "  %d,", bgx[i]-((i)<<4));
    if ((i&15)==15) fprintf(stdout, "\n");
  }
  fprintf(stdout, "};\n");
#endif
  fprintf(stdout, "static const unsigned char bgy[1024] = {\n");
  for (i = 0; i < 1024; i++) {
    int flagged = bgy[i] & (1<<30); bgy[i] &= ~(1<<30);
    fprintf(stdout, "  %d,", bgy[i] | (flagged?128:0));
    if ((i&15)==15) fprintf(stdout, "\n");
  }
  fprintf(stdout, "};\n");

  fprintf(stdout, "static const unsigned char fromx[1024] = {\n"); // in units of 1/16th of a seg, rel to seg start
  for (i = 0; i < 1024; i++) {
    int flagged = fromx[i] & (1<<30);
    fprintf(stdout, "  %d,", (fromx[i] & ~(1<<30)) | (flagged?128:0));
    if ((i&15)==15) fprintf(stdout, "\n");
  }
  fprintf(stdout, "};\n");
  fprintf(stdout, "static const unsigned char fromy[1024] = {\n");
  for (i = 0; i < 1024; i++) {
    int flagged = fromx[i] & (1<<30);
    fprintf(stdout, "  %d,", (fromy[i] & ~(1<<30)) | (flagged?128:0));
    if ((i&15)==15) fprintf(stdout, "\n");
  }
  fprintf(stdout, "};\n");
  fprintf(stdout, "static const unsigned char tox[1024] = {\n"); // in units of 1/16th of a seg, rel to seg start
  for (i = 0; i < 1024; i++) {
    int flagged = fromx[i] & (1<<30);
    fprintf(stdout, "  %d,", (tox[i] & ~(1<<30)) | (flagged?128:0));
    if ((i&15)==15) fprintf(stdout, "\n");
  }
  fprintf(stdout, "};\n");
  fprintf(stdout, "static const unsigned char toy[1024] = {\n");
  for (i = 0; i < 1024; i++) {
    int flagged = fromx[i] & (1<<30);
    fprintf(stdout, "  %d,", (toy[i] & ~(1<<30)) | (flagged?128:0));
    if ((i&15)==15) fprintf(stdout, "\n");
  }
  fprintf(stdout, "};\n");
  exit(0);  
}
