// experiments is 3d -> 2d projection
// Even 1 divide is too expensive.

#include <vectrex.h>
#include "controller.h"

static inline void set_scale(unsigned int scale) {
  /*VIA_t1_cnt_lo*/ *(volatile unsigned char *)0xD004 = scale;
}

#include "mountains.h"

static void SHOW_NUM(char *dest, long num) { // left-aligned
#define show_digit(x) *dest++ = (char)((x)+'0')
  int digit, zeroes;

  // This replaces code that used divide by 10 and modulo 10.  Much faster.

  // handles full 16 bit range of -32768:32767  -  Uses negative numbers to avoid the issue of negating -32768

  if (num >= 0) num = -num; else *dest++ = '-';
  digit = 0;
  zeroes = 1; // CLRing is shorter
  // max 11 add/subtracts...
  if (num <= -20000) { num += 20000; digit += 2; zeroes = 0; }
  if (num <= -10000) { num += 10000; digit += 1; zeroes = 0; }
  if (!zeroes) show_digit(digit);
  digit = 0;
  if (num <= -8000) { num += 8000; digit += 8; zeroes = 0; } else if (num <= -4000) { num += 4000; digit += 4; zeroes = 0; }
  if (num <= -2000) { num += 2000; digit += 2; zeroes = 0; }
  if (num <= -1000) { num += 1000; digit += 1; zeroes = 0; }
  if (!zeroes) show_digit(digit);
  digit = 0;
  if (num <= -800) { num += 800; digit += 8; zeroes = 0; } else if (num <= -400) { num += 400; digit += 4; zeroes = 0; }
  if (num <= -200) { num += 200; digit += 2; zeroes = 0; }
  if (num <= -100) { num += 100; digit += 1; zeroes = 0; }
  if (!zeroes) show_digit(digit);
  digit = 0;
  if (num <= -80) { num += 80; digit += 8; zeroes = 0; } else if (num <= -40) { num += 40; digit += 4; zeroes = 0; }
  if (num <= -20) { num += 20; digit += 2; zeroes = 0; }
  if (num <= -10) { num += 10; digit += 1; zeroes = 0; }
  if (!zeroes) show_digit(digit);
  show_digit((int)-num);
  *dest++ = ' '; *dest = 0x80;
}


// this is quite a "C" optimized version of print_sync
// (-O3 and no_frame_pointer)
#define ZERO_DELAY 5

void draw_synced_list_c(
	const signed char *u,
	signed int y,
	signed int x,
	unsigned int scaleMove,
	unsigned int scaleList
	)
{
#ifndef LINUX
	
	do
	{
		// resync / startsync
		dp_VIA_shift_reg = 0;		// all output is BLANK
		
		// move to zero
		dp_VIA_cntl = (unsigned int)0xcc;	// zero the integrators
		dp_VIA_port_a = 0;			// reset integrator offset
		dp_VIA_port_b = (int)0b10000010;
		
		dp_VIA_t1_cnt_lo = scaleMove;
		// delay, till beam is at zero
		// volatile - otherwise delay loop does not work with -O
		for (volatile signed int b=ZERO_DELAY; b>0; b--);
		dp_VIA_port_b= (int)0b10000011;
		
		// move to "location"
		dp_VIA_port_a = y;			// y pos to dac
		dp_VIA_cntl = (unsigned int)0xce;	// disable zero, disable all blank
		dp_VIA_port_b = 0;			// mux enable, dac to -> integrator y (and x)
		dp_VIA_port_b = 1;			// mux disable, dac only to x
		dp_VIA_port_a = x;			// dac -> x
		dp_VIA_t1_cnt_hi=0;		// start timer
		
		// this can be done before the wait loop
		// since it only fills the latch, not the actual timer!
		dp_VIA_t1_cnt_lo = scaleList;
		u+=3;
		
		// moveing test for yx== 0 into the move delay
		if ((*(u-2)!=0) || (*(u-1)!=0))
		{
			while ((dp_VIA_int_flags & 0x40) == 0); // wait till timer finishes
			
			// internal moveTo
			dp_VIA_port_a = *(u-2);	// y pos to dac
			dp_VIA_cntl = (unsigned int)0xce;	// disable zero, disable all blank
			dp_VIA_port_b = 0;			// mux enable, dac to -> integrator y (and x)
			dp_VIA_port_b = 1;			// mux disable, dac only to x
			dp_VIA_port_a = *(u-1);	// dac -> x
			dp_VIA_t1_cnt_hi=0;		// start timer
			while ((dp_VIA_int_flags & 0x40) == 0); // wait till timer finishes
		}
		else
		{
			while ((dp_VIA_int_flags & 0x40) == 0); // wait till timer finishes
		}
		
		while (1)
		{
			if (*u<0) // draw line
			{
				// draw a vector
				dp_VIA_port_a = *(1+u);		// y pos to dac
				dp_VIA_port_b = 0;				// mux enable, dac to -> integrator y (and x)
				dp_VIA_port_b=1;				// mux disable, dac only to x
				dp_VIA_port_a = *(2+u);		// dac -> x
				dp_VIA_t1_cnt_hi=0;			// start timer
				dp_VIA_shift_reg = (unsigned int)0xff;	// draw complete line

				if (scaleList>10)
				{
					while ((dp_VIA_int_flags & 0x40) == 0); // wait till timer finishes
				}
				dp_VIA_shift_reg = 0;			// all output is BLANK
			}
			else if (*u == 0) // moveTo
			{
				if ((*(u+1)!=0) || (*(u+2)!=0))
				{
					// internal moveTo
					dp_VIA_port_a = *(1+u);	// y pos to dac
					dp_VIA_cntl = (unsigned int)0xce;	// disable zero, disable all blank
					dp_VIA_port_b = 0;			// mux enable, dac to -> integrator y (and x)
					dp_VIA_port_b =1;			// mux disable, dac only to x
					dp_VIA_port_a = *(2+u);	// dac -> x
					dp_VIA_t1_cnt_hi=0;		// start timer
					while ((dp_VIA_int_flags & 0x40) == 0); // wait till timer finishes
				}
			}
			else
			{
				break;
			}
			u+=3;
		}
	} while (*u != 2);
#endif
}

long compass=0L,frame=-1L;

void line(int x0, int y0, int x1, int y1, unsigned int c) {
  Reset0Ref();Intensity_7F();
  Moveto_d(y0,x0);
    Draw_Line_d((y1-y0)/2,(x1-x0)/2);
    Draw_Line_d((y1-y0)-(y1-y0)/2,(x1-x0)-(x1-x0)/2);
(void)c;
}

void Radar(int cx, int cy) {
  // This can also be a sprite, including the rotating hand
  (void)cx;(void)cy;
}

void FixedGraphics(void) {
  // things that will go into synched lists eventually
}

typedef struct VPoint
{
  int x,y,z;
} VPoint;

static const VPoint shp_hovercraft_128[] = {
  {-52,47,0}, {0,0,60},
  {0,0,60}, {59,-54,0},
  {59,-54,0}, {0,0,-60},
  {0,0,-60}, {-52,47,0},
  {-23,44,0}, {0,0,40},
  {0,0,40}, {26,-54,0},
  {26,-54,0}, {0,0,-40},
  {0,0,-40}, {-23,44,0},
  {0,0,40}, {55,-22,0},
  {55,-22,0}, {0,0,-40},
  {0,0,-40}, {0,0,40},
  {-24,-9,10}, {-24,-9,-11},
  {-52,47,0}, {-23,44,0},
  {0,0,60}, {0,0,40},
  {59,-54,0}, {26,-54,0},
  {0,0,-60}, {0,0,-40},
  {0,0,60}, {0,0,40},
  {59,-54,0}, {55,-22,0},
  {0,0,-60}, {0,0,-40},
  {0,0,40}, {-24,-9,10},
  {0,0,-40}, {-24,-9,-11},
  {-52,47,0}, {-24,-9,10},
  {-52,47,0}, {-24,-9,-11},
  {-45,5,-5}, {-100,12,-5},
  {-45,5,4}, {-100,12,4},
  {-48,11,-4}, {-97,24,-5},
  {-48,11,3}, {-97,24,4},
  {-100,12,-5}, {-100,12,4},
  {-97,24,-5}, {-97,24,4},
  {-100,12,-5}, {-97,24,-5},
  {-100,12,4}, {-97,24,4},
  {-45,5,-5}, {-45,5,4},
  {-48,11,-4}, {-48,11,3},
  {-45,5,-5}, {-48,11,-4},
  {-45,5,4}, {-48,11,3},
  {0,0,0}, {0,0,0}
};

static const VPoint shp_obstacle_0[] = {
// test cube. Every vertex has 3 links
#ifdef CONVERTED
                                                  {-8,74,-75},                             {44,-61,-75}, 
                                                  {-8,74,-75},                {7,-74,-75},
                          {-8,74,74},             {-8,74,-75},
                                                                {-45,59,-75}, {7,-74,-75},
                                                                {-45,59,-75},              {44,-61,-75},
             {44,-61,74},                                                                  {44,-61,-75},
{-45,59,74},                           {7,-74,74},
{-45,59,74},                                                    {-45,59,-75},
{-45,59,74}, {44,-61,74},
             {44,-61,74}, {-8,74,74},
                          {-8,74,74},  {7,-74,74},
                                       {7,-74,74},                            {7,-74,-75}, 
#else
//P2          P6         P3           P1         P8               P4         P5           P7
  {40,40,40},{40,40,-40},
  {40,40,40},            {40,-40,40},
  {40,40,40},                        {-40,40,40},

                                                 {-40,-40,-40}, {-40,-40,40},
                                                 {-40,-40,-40},              {-40,40,-40},
                                                 {-40,-40,-40},                           {40,-40,-40},

                                     {-40,40,40},               {-40,-40,40},
                                     {-40,40,40},                            {-40,40,-40},

                         {40,-40,40},                           {-40,-40,40},
                         {40,-40,40},                                                     {40,-40,-40},
             {40,40,-40},                                                    {-40,40,-40},
             {40,40,-40},                                                                 {40,-40,-40},
#endif
{0,0,0}, {0,0,0},
};

typedef union {
struct { int High, Low; } Byte;
struct { long Word; } Long;
} Data;

void Project(long int x, long int y, long int z, long int *sx, long int *sy) {
  Data product;
  // z = -127 (near) to 127 (far)
  z = 127-(z>>2); // z = 254 to 0
  // when z = 254 want larger
  // when z = 0 want smallest
  product.Long.Word = x*z;
  *sx = product.Byte.Low>>1;
  product.Long.Word = y*z;
  *sy = product.Byte.Low>>1;
}

void Draw3D(VPoint *line) {
  VPoint *p1p, *p2p;
#define p1 (*p1p)
#define p2 (*p2p)
  long int lsx1,lsy1,lsx2,lsy2;
  int sx1,sy1,sx2,sy2;
  for (;;) {
    p1p = line++; p2p = line++;
    if ((p1.x|p1.y|p1.z|p2.x|p2.y|p2.z) == 0) break;
    Project(p1.x,p1.y,p1.z, &lsx1, &lsy1); sx1=(int)lsx1;sy1=(int)lsy1;
    Project(p2.x,p2.y,p2.z, &lsx2, &lsy2); sx2=(int)lsx2;sy2=(int)lsy2;
    //line(sx1, sy1, sx2, sy2, 127);
    Reset0Ref();Intensity_7F();
    Moveto_d(sy1,sx1);
    Draw_Line_d((sy2-sy1)/2,(sx2-sx1)/2);
    Draw_Line_d((sy2-sy1)-(sy2-sy1)/2,(sx2-sx1)-(sx2-sx1)/2);
  }
}


#include "vtank.h"
const signed char const *tank[128] = {
tank0, tank2, tank4, tank6, tank8, tank10, tank12, tank14, tank16, tank18, tank20, tank22, tank24, tank26, tank28, tank30,
tank32, tank34, tank36, tank38, tank40, tank42, tank44, tank46, tank48, tank50, tank52, tank54, tank56, tank58, tank60, tank62,
tank64, tank66, tank68, tank70, tank72, tank74, tank76, tank78, tank80, tank82, tank84, tank86, tank88, tank90, tank92, tank94,
tank96, tank98, tank100, tank102, tank104, tank106, tank108, tank110, tank112, tank114, tank116, tank118, tank120, tank122, tank124, tank126,
tank128, tank130, tank132, tank134, tank136, tank138, tank140, tank142, tank144, tank146, tank148, tank150, tank152, tank154, tank156, tank158,
tank160, tank162, tank164, tank166, tank168, tank170, tank172, tank174, tank176, tank178, tank180, tank182, tank184, tank186, tank188, tank190,
tank192, tank194, tank196, tank198, tank200, tank202, tank204, tank206, tank208, tank210, tank212, tank214, tank216, tank218, tank220, tank222,
tank224, tank226, tank228, tank230, tank232, tank234, tank236, tank238, tank240, tank242, tank244, tank246, tank248, tank250, tank252, tank254,
};

static int tweak = 0;
static int YOFFSET = 0;
void Mountains(void) {
  int x, prevx, prevy;
  long seg;

  // draw foreground mountains first:
  //set_scale(127);
  set_scale(200);
  Intensity_a(110);
  Reset0Ref();
  Moveto_d(0,0);
  Draw_Line_d(0,-112);
  Reset0Ref();
  Moveto_d(0,0);
  Draw_Line_d(0,112);

  seg = (compass-8)&1023;
  Reset0Ref(); Moveto_d(fgy[seg], tweak-111);
  Intensity_a(90);
  for (x = 1; x <15; x++) {
    prevy=fgy[seg];
    seg = (seg+1)&1023;    // draw 16 segments of 16 units wide
    if (prevy==0 && fgy[seg]==0) Moveto_d(fgy[seg]-prevy, 16); else Draw_Line_d(fgy[seg]-prevy, 16);
  }

//#ifdef NEVER
// Now background
#define BGY(n) (int)((bgy[n]&127))
#define FLAGGED(n) ((int)(fromx[n]&128))

  //set_scale(127);
  //set_scale(145);
  seg = (compass-8)&1023;
  Intensity_a(55);
  for (x=0;x<14;x++) {

    if (!FLAGGED(seg)) {
      int from_x, from_y,to_x,to_y;

      // These are relative to the left edge of this cell
      from_x = (int)(fromx[seg]&127);
      from_y = (int)(fromy[seg]&127);
      to_x = (int)(tox[seg]&127);
      to_y = (int)(toy[seg]&127);

      Reset0Ref();
      // Abs move
      Moveto_d(from_y,from_x - 111 + tweak+ x*16);
      prevx=from_x; prevy=from_y;

      // Rel draw
      Draw_Line_d(to_y-from_y, to_x-from_x );
    }
    seg=(seg+1)&1023L;
  }
//#endif
}

int jx, jy;
#define jTHRESH 0x20
int main(void) {
  unsigned int i=96;
  //const int clockwise = 1;
  //unsigned int mag=20;
  //int x=0,y = -4;
  (void)i;
  // for (mag = 20; mag < 255; mag++) {

  YOFFSET=6;
  enable_controller_1_x();
  enable_controller_1_y();
  compass=559;
  for (;;) {
    frame++;

    Wait_Recal ();
    Reset0Ref ();
    Intensity_7F ();
    set_scale (127);
    {
      static char num[10];
      SHOW_NUM(num, compass); // left-aligned
      Print_Str_d ( 124,-24, num);
    }
    check_buttons();
    check_joysticks();
    set_scale(64);
    Draw3D((VPoint *)&shp_obstacle_0[0]);
    FixedGraphics();
    jx = joystick_1_x();
    if (jx < 0) {
        tweak+=1;
        if (tweak==16) {
          tweak=0;
          compass--;
        }
    } else if (jx > 0) {
        tweak -= 1;
        if (tweak==-16) {
          tweak=0;
          compass++;
        }
    }
    jy = joystick_1_y();
    if (jy < 0) {
      if ((frame&3)==3) {
        YOFFSET -= 1;
      }
    } else if (jy > 0) {
      if ((frame&3)==3) {
        YOFFSET += 1;
      }
    }
    compass&=1023;
    // Mountains();
    Reset0Ref();
    Intensity_a(50);
    //draw_synced_list_c(tank[clockwise?95:31], y, x,  127,mag);

    if (1||(frame&1)==1) {
    if (1) {
      tweak -= 1;
      if (tweak==-16) {
        tweak=0;
        compass++;
      }
    } else if (0) {
      tweak += 1;
      if (tweak==16) {
        tweak=0;
        compass--;
      }
    }
    }
  }

  return 0;
}
